package org.siforge.mvc;


import org.apache.log4j.*;
import java.io.Serializable;
import java.sql.*;
import java.sql.Connection;
import java.util.*;
import java.util.List;


/** Questa lista e' usata dal model per interfacciarsi in modo efficace con la view.
 * Modella una lista di option bean chiave,valore (due stringhe).
 * E' possibile accedere alla chiave e ottenere il valore
 * ed e' parimenti possibile ottenere la lista ordinata.
 * <br/>
 * Per semplicita' questa classe ricostruisce i suoi indici ogni volta che si aggiungono gruppi di elementi o si tolgono elementi.
 * E' ottimizzata solo l'aggiunta di un singolo elemento, perche' tanto gli altri non sono usati.
 * BUG: Se vengono inserite chiavi duplicate o erronee il comportamento e' indefinito. 
 * 
 * Il codice tenta di lanciare un'eccezione in questi  casi, ma non e' detto che li rilevi sempre.
 * 
 * Esempio: se si inseriscono le coppie:
 * <pre>
 * Name, Value
 * "desc1", "errore"
 * "desc2", "errore"
 * </pre>
 * @author Giovanni Giorgi
 * @version 2.0
 * @vompany SIForge.org
 */
public class OptionsList extends  ArrayList implements java.io.Serializable {


    /** Indici*/
    private TreeMap name2Option, value2Option;

    public OptionsList(){
		super();
		name2Option=new TreeMap();
		value2Option= new TreeMap();		
    }

    /** Fallisce se l'array list non e' una lista di OptionBean
     */
    public static OptionsList fromArrayList(ArrayList r){
		OptionsList fl=new OptionsList();
		fl.addAll(r);
		return fl;
    }

    public static  OptionsList fromSQL(Connection conn, 
									   String tableName, String codeRowName, String descRowName)
		throws SQLException 
    {
		return fromSQL(conn,tableName,codeRowName,descRowName, null,null);
    }
	
    public static OptionsList fromSQL(Connection conn, 
									  String tableName, String codeRowName, String descRowName,  String orderby)
		throws SQLException
    {
		return fromSQL(conn,tableName,codeRowName,descRowName,orderby,null);
    }
	
    /** orderby e whereFilter possono essere a null*/
    public static  OptionsList fromSQL(Connection conn, String tableName, 
                                       String codeRowName, 
                                       String descRowName, 
                                       String orderby, String whereFilter)
		throws SQLException
    {
		PreparedStatement stm=null;
		OptionsList aList = null;
		long tStart=System.currentTimeMillis();
		String query="SELECT "+ codeRowName+", "+ descRowName+" FROM "+tableName;
		if(whereFilter!=null){
			query += " WHERE "+whereFilter;
		}
		if(orderby!=null){
			query += " ORDER BY "+orderby;
		}
		Logger log=Logger.getLogger("OptionsList.static");
		ResultSet rs;
		try {
			stm = conn.prepareStatement(query);
			rs = stm.executeQuery();
		} catch (SQLException e) {
			log.error("Cannot execute:" + stm, e);
			throw e;
		}
		aList= new OptionsList();
		while (rs.next()) {
			// L'assegnamento e' sempre descrizione, codice.
			aList.addOption( rs.getString(2), rs.getString(1) );
		}
		log.debug(System.currentTimeMillis() -tStart +" ms OptionsList.fromSQL <--< "+stm);
		return aList;				
    }
	


    private void rebuildTreeMaps(){
		
		name2Option=new TreeMap();
		value2Option= new TreeMap();
		
		Object c[]=super.toArray();
		
		for(int i=0; i < c.length; i++){
			OptionBean  elem=(OptionBean) c[i];
			safeUpdateIndexes(elem);
		}
    }

    private void safeUpdateIndexes(OptionBean elem)
    {
		if(value2Option.containsKey(elem.getValue())){
			OptionBean duplicated=(OptionBean) value2Option.get(elem.getValue());
			String msg="Duplicated key! Elem="+elem+" has a value already stored as="+duplicated;
			throw new IllegalArgumentException(msg);
		}
		name2Option.put(elem.getName(),  elem);
		value2Option.put(elem.getValue(), elem);
    }
	


    /** Effettua automaticamente l'autotrim. usando come implementazione il SimpleOptionBean
     *  I parametri sono
     */
    public boolean addOption(String descrizioneLunga, String codiceDB) {
		return this.add(new SimpleOptionBean(descrizioneLunga, codiceDB));
    }

    /** Ritorna null se l'elemento non e' presente
     * Dato un nome "lungo" ritorna un "nome breve" (=un value select o un indice univoco)
     */		
    public String getValue(String longName) {
		if(!name2Option.containsKey(longName)) { return null; }
		return ((OptionBean)name2Option.get(longName)).getValue();
    }
	
		
    /** Ritorna null se l'elemento non e' presente
     */
    public String getName(String valueKey){
		if(!value2Option.containsKey(valueKey)){ return null; }
		return ((OptionBean)value2Option.get(valueKey)).getName();
		
    }


	
	
    /** Deve ritornare -1 se l'oggetto non e' presente
     */
    public int indexOfName(String key){
		// Trova l'oggetto:
		if(! name2Option.containsKey(key)){
			return -1;
		}
		OptionBean o=(OptionBean) name2Option.get(key);
		return super.indexOf(o);
    }
	
    /** Deve ritornare -1 se il valore non e' presente
     */
    public int indexOfValue(String v){
		if(! value2Option.containsKey(v)){
			return -1;
		}
		OptionBean o=(OptionBean) value2Option.get(v);
		return super.indexOf(o);
    }
	

    public String toString(){
		if(size()<=25){
			return "[ FL("+name2Option.size()+") "+super.toString()+" ]";
		}else{
			// Stampo meno roba:
			StringBuffer b=new StringBuffer();
			b.append("[ FL("+name2Option.size()+") ");
			b.append( this.get(0)); b.append(" , ");
			b.append( this.get(1));
			b.append(" ... ");
			b.append( this.get(size()-2)); b.append(" , ");
			b.append( this.get(size()-1));
			return b.toString();
		}
    }		
	
	
    /** Metodo di add ottimizzato
     * @see java.util.Collection#add(Object)
     */
    public boolean add(Object o) {
		OptionBean elem;
		
		// Il codice che usava instanceof non va sempre!!
		// Bug di jdk di ibm?[GG]
		try {			
			elem=(OptionBean)o;
		} catch (ClassCastException e) {
			// e.printStackTrace();
			throw new IllegalArgumentException("Fast List can manage only OptionBean-s interfaces. Invalid Obj="+o);
		}
		
		boolean resp=super.add(elem);
		// Aggiorna le sue due treemap:
		
		safeUpdateIndexes(elem);
		return resp;
    }


    /** add con indice
     * @see java.util.List#add(int, Object)
     */

    public void add(int index, Object element) {
		if(!(element instanceof OptionBean)){
			throw new IllegalArgumentException("add(int,element): Fast List can manage only OptionsList.OptionBean-s interfaces");
		}
		super.add(index, element);
		rebuildTreeMaps();
    }


    /**
     * @see java.util.Collection#addAll(Collection)
     */
    public boolean addAll(Collection c) {
		boolean resp=super.addAll(c);
		rebuildTreeMaps();
		return resp;
    }

    /**
     * @see java.util.List#addAll(int, Collection)
     */
    public boolean addAll(int index, Collection c) {
		boolean resp=super.addAll(index, c);
		rebuildTreeMaps();
		return resp;
    }

    /**
     * @see java.util.List#remove(int)
     */
    public Object remove(int index) {
		Object resp=super.remove(index);
		rebuildTreeMaps();
		return resp;
    }

    /**
     * @see java.util.AbstractList#removeRange(int, int)
     */
    protected void removeRange(int fromIndex, int toIndex) {
		super.removeRange(fromIndex, toIndex);
		rebuildTreeMaps();
    }


    /**
     * @see java.util.Collection#clear()
     */
    public void clear() {
		super.clear();
		// Ok, basterebbe rimuoverli ma perche' scrivere piu' codice?
		rebuildTreeMaps();
    }


    /**
     * @see java.util.Collection#remove(Object)
     */
    public boolean remove(Object o) {
		boolean resp=super.remove(o);
		rebuildTreeMaps();
		return resp;
    }

    /**
     * @see java.util.Collection#removeAll(Collection)
     */
    public boolean removeAll(Collection c) {
		boolean resp=super.removeAll(c);
		rebuildTreeMaps();
		return resp;
    }

    /**
     * @see java.util.Collection#retainAll(Collection)
     */
    public boolean retainAll(Collection c) {
		boolean resp=super.retainAll(c);
		rebuildTreeMaps();
		return resp;
    }


    /** 
     * @see java.util.List#set(int, Object)
     */
    public Object set(int index, Object element) {
		Object obj=super.set(index, element);
		rebuildTreeMaps();
		return obj;
    }

}
